@@ -1,4 +1,4 @@
-/*  k4start.c,v 1.7 2005/05/01 18:41:22 eagle Exp
+/*  $Id: k4start.c 2348 2006-01-02 01:43:25Z rra $
 **
 **  Kerberos v4 kinit replacement suitable for daemon authentication.
 **
@@ -27,24 +27,43 @@
 #include <stdlib.h>
 #include <string.h>
 #include <sys/stat.h>
-#include <sys/time.h>
 #include <sys/types.h>
-#include <time.h>
 #include <unistd.h>
 
+#if TIME_WITH_SYS_TIME
+# include <sys/time.h>
+# include <time.h>
+#else
+# if HAVE_SYS_TIME_H
+#  include <sys/time.h>
+# else
+#  include <time.h>
+# endif
+#endif
+
+#if HAVE_K_SETPAG
+# if HAVE_KAFS_H
+#  include <kafs.h>
+# endif
+# define lsetpag() k_setpag()
+#elif HAVE_LSETPAG
+int lsetpag(void);
+#else
+# define lsetpag() (0)
+#endif
+
 #ifdef HAVE_KERBEROSIV_KRB_H
 # include <kerberosIV/krb.h>
 #else
 # include <krb.h>
 #endif
 
-#ifndef HAVE_MKSTEMP
-extern int mkstemp(char *);
+#ifndef HAVE_DAEMON
+extern int daemon(int, int);
 #endif
 
-/* The AFS headers don't prototype this. */
-#ifdef HAVE_SETPAG
-int setpag(void);
+#ifndef HAVE_MKSTEMP
+extern int mkstemp(char *);
 #endif
 
 /* We default to a ten hour ticket lifetime if the Kerberos headers don't
@@ -93,6 +112,7 @@ Usage: k4start [options] [name]\n\
    -I <service instance>        (default: realm name)\n\
    -r <service realm>           (default: local realm)\n\
 \n\
+   -b                   Fork and run in the background\n\
    -f <srvtab>          Read password from <srvtab>, as a srvtab key\n\
    -H <limit>           Check for a happy ticket, one that doesn't expire in\n\
                         less than <limit> minutes, and exit 0 if it's okay,\n\
@@ -102,6 +122,7 @@ Usage: k4start [options] [name]\n\
    -k <file>            Use <file> as the ticket cache\n\
    -l <lifetime>        Ticket lifetime in minutes\n\
    -n                   Don't run aklog or KINIT_PROG\n\
+   -p <file>            Write process ID (PID) to <file>\n\
    -q                   Don't output any unnecessary text\n\
    -s                   Read password on standard input\n\
    -t                   Get AFS token via aklog or KINIT_PROG\n\
@@ -228,10 +249,12 @@ int
 main(int argc, char *argv[])
 {
     struct options options;
-    int k_errno, option, result;
+    int k_errno, opt, result;
     char *username = NULL;
     char *aklog = NULL;
     char **command = NULL;
+    char *pidfile = NULL;
+    int background = 0;
     int lifetime = DEFAULT_TKT_LIFE;
     pid_t child = 0;
     int status = 0;
@@ -239,10 +262,12 @@ main(int argc, char *argv[])
 
     /* Parse command-line options. */
     memset(&options, 0, sizeof(options));
-    while ((option = getopt(argc, argv, "f:H:I:i:K:k:l:npqr:S:stu:v")) != EOF)
-        switch (option) {
+    while ((opt = getopt(argc, argv, "bf:H:I:i:K:k:l:np:qr:S:stu:v")) != EOF)
+        switch (opt) {
+        case 'b': background = 1;               break;
         case 'k': options.cache = optarg;       break;
         case 'n': options.no_aklog = 1;         break;
+        case 'p': pidfile = optarg;             break;
         case 'q': options.quiet = 1;            break;
         case 't': options.run_aklog = 1;        break;
         case 'v': options.verbose = 1;          break;
@@ -296,7 +321,6 @@ main(int argc, char *argv[])
                 die("service name %s too long (%lu max)", optarg,
                     (unsigned long) sizeof(options.sname));
             break;
-        case 'p':
         case 's':
             options.stdin_passwd = 1;
             if (options.srvtab != NULL)
@@ -322,6 +346,10 @@ main(int argc, char *argv[])
         command = argv + 1;
 
     /* Check the arguments for consistency. */
+    if (background && options.srvtab == NULL)
+        die("-b option requires a srvtab be specified with -f");
+    if (background && options.keep_ticket == 0 && command == NULL)
+        die("-b only makes sense with -K or a command to run");
     if (options.keep_ticket > 0 && options.srvtab == NULL)
         die("-K option requires a srvtab be specified with -f");
     if (command != NULL && options.srvtab == NULL)
@@ -359,7 +387,7 @@ main(int argc, char *argv[])
         int fd;
 
         options.cache = malloc(strlen("/tmp/tkt_XXXXXX") + 20 + 1);
-        sprintf(options.cache, "/tmp/tkt%d_XXXXXX", getuid());
+        sprintf(options.cache, "/tmp/tkt%d_XXXXXX", (int) getuid());
         fd = mkstemp(options.cache);
         if (fd < 0)
             die("cannot create ticket cache file: %s", strerror(errno));
@@ -378,9 +406,9 @@ main(int argc, char *argv[])
         putenv(env);
     }
 
-    /* If either -K or -H were given, set quiet automatically unless verbose
-       was set. */
-    if (options.keep_ticket > 0 || options.happy_ticket > 0)
+    /* If -K, -H, or -b were given, set quiet automatically unless verbose was
+       set. */
+    if (options.keep_ticket > 0 || options.happy_ticket > 0 || background)
         if (!options.verbose)
             options.quiet = 1;
 
@@ -431,15 +459,33 @@ main(int argc, char *argv[])
 
     /* If built with setpag support and we're running a command, create the
        new PAG now before the first authentication. */
-#ifdef HAVE_SETPAG
     if (command != NULL && options.run_aklog && !options.no_aklog)
-        if (setpag() < 0)
+        if (lsetpag() < 0)
             die("unable to create PAG: %s", strerror(errno));
-#endif
 
-    /* Now, do the actual authentication and then run our child command, if
-       one was specified. */
+    /* Now, do the actual authentication. */
     status = authenticate(&options, aklog);
+
+    /* If told to background, background ourselves.  We do this late so that
+       we can report initial errors.  We have to do this before spawning the
+       command, though, since we want to background the command as well and
+       since otherwise we wouldn't be able to wait for the child process. */
+    if (background)
+        daemon(0, 0);
+
+    /* Write out the PID file.  Note that we can't report failures usefully,
+       since this is generally used with -b. */
+    if (pidfile != NULL) {
+        FILE *file;
+
+        file = fopen(pidfile, "w");
+        if (file != NULL) {
+            fprintf(file, "%lu\n", (unsigned long) getpid());
+            fclose(file);
+        }
+    }
+
+    /* Spawn the external command, if we were told to run one. */
     if (command != NULL) {
         child = start_command(command[0], command);
         if (child < 0)
