@@ -1,9 +1,9 @@
-/*  k5start.c,v 1.25 2005/06/04 20:21:01 eagle Exp
+/*  $Id: k5start.c 2348 2006-01-02 01:43:25Z rra $
 **
 **  Kerberos v5 kinit replacement suitable for daemon authentication.
 **
 **  Copyright 1987, 1988 by the Massachusetts Institute of Technology.
-**  Copyright 1995, 1996, 1997, 1999, 2000, 2001, 2002, 2004, 2005
+**  Copyright 1995, 1996, 1997, 1999, 2000, 2001, 2002, 2004, 2005, 2006
 **      Board of Trustees, Leland Stanford Jr. University
 **
 **  For copying and distribution information, please see README.
@@ -31,18 +31,56 @@
 #include <string.h>
 #include <sys/stat.h>
 #include <sys/types.h>
-#include <time.h>
 #include <unistd.h>
 
+#if TIME_WITH_SYS_TIME
+# include <sys/time.h>
+# include <time.h>
+#else
+# if HAVE_SYS_TIME_H
+#  include <sys/time.h>
+# else
+#  include <time.h>
+# endif
+#endif
+
 #include <krb5.h>
 
-#ifndef HAVE_MKSTEMP
-extern int mkstemp(char *);
+#if HAVE_K_SETPAG
+# if HAVE_KAFS_H
+#  include <kafs.h>
+# endif
+# define lsetpag() k_setpag()
+#elif HAVE_LSETPAG
+int lsetpag(void);
+#else
+# define lsetpag() (0)
 #endif
 
-/* The AFS headers don't prototype this. */
-#ifdef HAVE_SETPAG
-int setpag(void);
+/* __attribute__ is available in gcc 2.5 and later, but only with gcc 2.7
+   could you use the __format__ form of the attributes, which is what we use
+   (to avoid confusion with other macros). */
+#ifndef __attribute__
+# if __GNUC__ < 2 || (__GNUC__ == 2 && __GNUC_MINOR__ < 7)
+#  define __attribute__(spec)   /* empty */
+# endif
+#endif
+
+#ifndef HAVE_DAEMON
+extern int daemon(int, int);
+#endif
+
+#ifndef HAVE_KRB5_ERR
+extern krb5_error_code krb5_err(krb5_context, int, krb5_error_code,
+                                const char *, ...)
+    __attribute__((__format__(printf, 4, 5)));
+extern krb5_error_code krb5_warn(krb5_context, krb5_error_code,
+                                 const char *, ...)
+    __attribute__((__format__(printf, 3, 4)));
+#endif
+
+#ifndef HAVE_MKSTEMP
+extern int mkstemp(char *);
 #endif
 
 /* The default ticket lifetime in minutes.  Default to 10 hours. */
@@ -87,6 +125,7 @@ Usage: k5start [options] [name [command]]\n\
    -I <service instance>        (default: realm name)\n\
    -r <service realm>           (default: local realm)\n\
 \n\
+   -b                   Fork and run in the background\n\
    -f <keytab>          Use <keytab> for authentication rather than password\n\
    -H <limit>           Check for a happy ticket, one that doesn't expire in\n\
                         less than <limit> minutes, and exit 0 if it's okay,\n\
@@ -95,6 +134,7 @@ Usage: k5start [options] [name [command]]\n\
                         (implies -q unless -v is given)\n\
    -k <file>            Use <file> as the ticket cache\n\
    -l <lifetime>        Ticket lifetime in minutes\n\
+   -p <file>            Write process ID (PID) to <file>\n\
    -q                   Don't output any unnecessary text\n\
    -s                   Read password on standard input\n\
    -t                   Get AFS token via aklog or KINIT_PROG\n\
@@ -216,7 +256,7 @@ authenticate(krb5_context ctx, struct options *options)
 
         k5_errno = krb5_unparse_name(ctx, options->kprinc, &p);
         if (k5_errno != 0)
-            com_err("k5start", k5_errno, "when unparsing name");
+            krb5_warn(ctx, k5_errno, "k5start: error unparsing name");
         else {
             printf("Principal: %s\n", p);
             free(p);
@@ -225,11 +265,9 @@ authenticate(krb5_context ctx, struct options *options)
     }
     if (options->keytab != NULL) {
         k5_errno = krb5_kt_resolve(ctx, options->keytab, &k5_keytab);
-        if (k5_errno != 0) {
-            com_err("k5start", k5_errno, "resolving keytab %s",
-                    options->keytab);
-            exit(1);
-        }
+        if (k5_errno != 0)
+            krb5_err(ctx, 1, k5_errno, "k5start: error resolving keytab %s",
+                     options->keytab);
         k5_errno = krb5_get_init_creds_keytab(ctx, &creds,
                                               options->kprinc, k5_keytab,
                                               0, options->service,
@@ -253,24 +291,18 @@ authenticate(krb5_context ctx, struct options *options)
             die("password too long");
         k5_errno = krb5_get_init_creds_password(ctx, &creds,
                                                 options->kprinc, buffer,
-                                                krb5_prompter_posix, NULL,
-                                                0, options->service,
+                                                NULL, NULL, 0,
+                                                options->service,
                                                 &options->kopts);
     }
-    if (k5_errno != 0) {
-        com_err("k5start", k5_errno, "while getting initial credentials");
-        exit(1);
-    }
+    if (k5_errno != 0)
+        krb5_err(ctx, 1, k5_errno, "k5start: error getting credentials");
     k5_errno = krb5_cc_initialize(ctx, options->ccache, options->kprinc);
-    if (k5_errno != 0) {
-        com_err("k5start", k5_errno, "while initializing ticket cache");
-        exit(1);
-    }
+    if (k5_errno != 0)
+        krb5_err(ctx, 1, k5_errno, "k5start: error initializing ticket cache");
     k5_errno = krb5_cc_store_cred(ctx, options->ccache, &creds);
-    if (k5_errno != 0) {
-        com_err("k5start", k5_errno, "while storing credentials");
-        exit(1);
-    }
+    if (k5_errno != 0)
+        krb5_err(ctx, 1, k5_errno, "k5start: error storing credentials");
 
     /* Make sure that we don't free princ; we use it later. */
     if (creds.client == options->kprinc)
@@ -304,22 +336,17 @@ first_principal(krb5_context ctx, const char *path)
     char *principal = NULL;
 
     k5_errno = krb5_kt_resolve(ctx, path, &keytab);
-    if (k5_errno != 0) {
-        com_err("k5start", k5_errno, "while opening %s", path);
-        exit(1);
-    }
+    if (k5_errno != 0)
+        krb5_err(ctx, 1, k5_errno, "k5start: error opening %s", path);
     k5_errno = krb5_kt_start_seq_get(ctx, keytab, &cursor);
-    if (k5_errno != 0) {
-        com_err("k5start", k5_errno, "while reading %s", path);
-        exit(1);
-    }
+    if (k5_errno != 0)
+        krb5_err(ctx, 1, k5_errno, "k5start: error reading %s", path);
     k5_errno = krb5_kt_next_entry(ctx, keytab, &entry, &cursor);
     if (k5_errno == 0) {
         k5_errno = krb5_unparse_name(ctx, entry.principal, &principal);
-        if (k5_errno != 0) {
-            com_err("k5start", k5_errno, "while unparsing name from %s", path);
-            exit(1);
-        }
+        if (k5_errno != 0)
+            krb5_err(ctx, 1, k5_errno,
+                     "k5start: error unparsing name from %s", path);
 #ifdef HAVE_KRB5_FREE_KEYTAB_ENTRY_CONTENTS
         krb5_free_keytab_entry_contents(ctx, &entry);
 #else
@@ -340,7 +367,7 @@ int
 main(int argc, char *argv[])
 {
     struct options options;
-    int k5_errno, option, result;
+    int k5_errno, opt, result;
     size_t length;
     const char *inst = NULL;
     const char *sname = NULL;
@@ -349,6 +376,8 @@ main(int argc, char *argv[])
     char *cache = NULL;
     char *principal = NULL;
     char **command = NULL;
+    char *pidfile = NULL;
+    int background = 0;
     int lifetime = DEFAULT_LIFETIME;
     krb5_context ctx;
     krb5_deltat life_secs;
@@ -359,12 +388,14 @@ main(int argc, char *argv[])
 
     /* Parse command-line options. */
     memset(&options, 0, sizeof(options));
-    while ((option = getopt(argc, argv, "f:H:I:i:K:k:l:npqr:S:stUu:v")) != EOF)
-        switch (option) {
+    while ((opt = getopt(argc, argv, "bf:H:I:i:K:k:l:np:qr:S:stUu:v")) != EOF)
+        switch (opt) {
+        case 'b': background = 1;               break;
         case 'I': sinst = optarg;               break;
         case 'i': inst = optarg;                break;
         case 'k': cache = optarg;               break;
         case 'n': /* Ignored */                 break;
+        case 'p': pidfile = optarg;             break;
         case 'q': options.quiet = 1;            break;
         case 'r': srealm = optarg;              break;
         case 'S': sname = optarg;               break;
@@ -394,7 +425,6 @@ main(int argc, char *argv[])
                 die("bad lifetime value %s, use 10h 10m format", optarg);
             lifetime = life_secs / 60;
             break;
-        case 'p':
         case 's':
             options.stdin_passwd = 1;
             if (options.keytab != NULL)
@@ -422,6 +452,10 @@ main(int argc, char *argv[])
         command = argv;
 
     /* Check the arguments for consistency. */
+    if (background && options.keytab == NULL)
+        die("-b option requires a keytab be specified with -f");
+    if (background && options.keep_ticket == 0 && command == NULL)
+        die("-b only makes sense with -K or a command to run");
     if (options.keep_ticket > 0 && options.keytab == NULL)
         die("-K option requires a keytab be specified with -f");
     if (command != NULL && options.keytab == NULL)
@@ -441,21 +475,12 @@ main(int argc, char *argv[])
     if (options.aklog == NULL && options.run_aklog)
         die("set KINIT_PROG to specify the path to aklog");
 
-    /* Heimdal initializes the error table using state stored in the context,
-       but alas this means com_err doesn't work right.  This hack seems to fix
-       it. */
-#ifdef HAVE_INITIALIZE_KRB5_ERROR_TABLE
-    initialize_krb5_error_table();
-#endif
-
     /* Establish a K5 context. */
     k5_errno = krb5_init_context(&ctx);
-    if (k5_errno != 0) {
-        com_err("k5start", k5_errno, "while initializing Kerberos 5 library");
-        exit(1);
-    }
+    if (k5_errno != 0)
+        krb5_err(ctx, 1, k5_errno, "k5start: error initializing Kerberos");
 
-    /* If the -P option was given, figure out the principal from the keytab. */
+    /* If the -U option was given, figure out the principal from the keytab. */
     if (search_keytab)
         principal = first_principal(ctx, options.keytab);
 
@@ -477,7 +502,7 @@ main(int argc, char *argv[])
         int fd;
 
         cache = malloc(strlen("/tmp/krb5cc__XXXXXX") + 20 + 1);
-        sprintf(cache, "/tmp/krb5cc_%d_XXXXXX", getuid());
+        sprintf(cache, "/tmp/krb5cc_%d_XXXXXX", (int) getuid());
         fd = mkstemp(cache);
         if (fd < 0)
             die("cannot create ticket cache file: %s", strerror(errno));
@@ -497,14 +522,12 @@ main(int argc, char *argv[])
         putenv(env);
         k5_errno = krb5_cc_resolve(ctx, cache, &options.ccache);
     }
-    if (k5_errno != 0) {
-        com_err("k5start", k5_errno, "while initializing ticket cache");
-        exit(1);
-    }
+    if (k5_errno != 0)
+        krb5_err(ctx, 1, k5_errno, "k5start: error initializing ticket cache");
 
-    /* If either -K or -H were given, set quiet automatically unless verbose
-       was set. */
-    if (options.keep_ticket > 0 || options.happy_ticket > 0)
+    /* If -K, -H, or -b were given, set quiet automatically unless verbose was
+       set. */
+    if (options.keep_ticket > 0 || options.happy_ticket > 0 || background)
         if (!options.verbose)
             options.quiet = 1;
 
@@ -524,10 +547,8 @@ main(int argc, char *argv[])
         principal = p;
     }
     k5_errno = krb5_parse_name(ctx, principal, &options.kprinc);
-    if (k5_errno != 0) {
-        com_err("k5start", k5_errno, "when parsing %s", principal);
-        exit(1);
-    }
+    if (k5_errno != 0)
+        krb5_err(ctx, 1, k5_errno, "k5start: error parsing %s", principal);
 
     /* Display the identity that we're obtaining Kerberos tickets for.  We do
        this by unparsing the principal rather than using username and inst
@@ -536,10 +557,9 @@ main(int argc, char *argv[])
         char *p;
 
         k5_errno = krb5_unparse_name(ctx, options.kprinc, &p);
-        if (k5_errno != 0) {
-            com_err("k5start", k5_errno, "when unparsing name %s", principal);
-            exit(1);
-        }
+        if (k5_errno != 0)
+            krb5_err(ctx, 1, k5_errno, "k5start: error unparsing name %s",
+                     principal);
         printf("Kerberos initialization for %s", p);
         free(p);
         if (sname != NULL) {
@@ -566,10 +586,9 @@ main(int argc, char *argv[])
     sprintf(options.service, "%s/%s@%s", sname, sinst, srealm);
     status = krb5_build_principal(ctx, &options.ksprinc, strlen(srealm),
                                   srealm, sname, sinst, (const char *) NULL);
-    if (status != 0) {
-        com_err("k5start", status, "while creating service principal name");
-        exit(1);
-    }
+    if (status != 0)
+        krb5_err(ctx, 1, status,
+                 "k5start: error creating service principal name");
 
     /* Figure out our ticket lifetime and initialize the options. */
     life_secs = lifetime * 60;
@@ -588,14 +607,33 @@ main(int argc, char *argv[])
 
     /* If built with setpag support and we're running a command, create the
        new PAG now before the first authentication. */
-#ifdef HAVE_SETPAG
     if (command != NULL && options.run_aklog)
-        if (setpag() < 0)
+        if (lsetpag() < 0)
             die("unable to create PAG: %s", strerror(errno));
-#endif
 
     /* Now, the actual authentication part. */
     status = authenticate(ctx, &options);
+
+    /* If told to background, background ourselves.  We do this late so that
+       we can report initial errors.  We have to do this before spawning the
+       command, though, since we want to background the command as well and
+       since otherwise we wouldn't be able to wait for the child process. */
+    if (background)
+        daemon(0, 0);
+
+    /* Write out the PID file.  Note that we can't report failures usefully,
+       since this is generally used with -b. */
+    if (pidfile != NULL) {
+        FILE *file;
+
+        file = fopen(pidfile, "w");
+        if (file != NULL) {
+            fprintf(file, "%lu\n", (unsigned long) getpid());
+            fclose(file);
+        }
+    }
+
+    /* Spawn the external command, if we were told to run one. */
     if (command != NULL) {
         child = start_command(command[0], command);
         if (child < 0)
